 /****************************************************************************
 Copyright 2013 Sony Corporation.

 Abstract:  Memory Allocator for Atheros Wlan Driver
 Author:    Tokuhiro Nishikawa
 Mail:      Tokuhiro.Nishikawa@jp.sony.com
****************************************************************************/

// ***************************************************************************
// include
// ***************************************************************************
#include <udif/module.h>
#include <udif/delay.h>
#include <udif/cache.h>
#include <udif/device.h>
#include <udif/malloc.h>

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/bug.h>
#include <asm/uaccess.h>

#include "wlanMemoryAllocator.h"

// ***************************************************************************
// Driver Descriptions
// ***************************************************************************
#define DRIVER_VERSION  "v1.0"
#define DRIVER_AUTHOR   "Sony Corporation"
#define DRIVER_DESC     "wlanMemoryAllocator"
#define DRIVER_LICENSE  "Proprietary"

// ***************************************************************************
// Module Information
// ***************************************************************************
MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE( DRIVER_LICENSE );

// ***************************************************************************
// Global Variables
// ***************************************************************************
static void * g_pMemGfpKernel[WMA_STATIC_ALLOCATION_NUM];
static int    g_allocateCounter;

// ***************************************************************************
// Internal Functions
// ***************************************************************************

//-------------------------------------------------------------
// initialization
// Allocate 128KB of memory.
//-------------------------------------------------------------
static int __init wlan_allocator_init(void)
{
    int ret = 0;
    int i = 0;

    for(i=0; i<WMA_STATIC_ALLOCATION_NUM; i++)
    {
        g_pMemGfpKernel[i] = kmalloc(WMA_STATIC_ALLOCATION_SIZE, GFP_KERNEL);
        if(NULL == g_pMemGfpKernel[i])
        {
            //retry
            g_pMemGfpKernel[i] = kmalloc(WMA_STATIC_ALLOCATION_SIZE, GFP_KERNEL);
            if(NULL == g_pMemGfpKernel[i])
            {
                // If allocation failed, its fatal. No chance of recovering.
                printk("wlanMemoryAllocator: [ERROR] Memory Allocation Failure");
                BUG_ON(!(NULL == g_pMemGfpKernel[i]));
            }
        }
        memset(g_pMemGfpKernel[i], 0, WMA_STATIC_ALLOCATION_SIZE);
        printk("wlanMemoryAllocator: Memory Allocated. g_pMemGfpKernel[%d] = %p, size = %d\n", i, g_pMemGfpKernel[i], WMA_STATIC_ALLOCATION_SIZE);
    }

    g_allocateCounter = 0;

    return ret;
}

//-------------------------------------------------------------
// Finalization
// Free the memory.
//-------------------------------------------------------------
static void __exit wlan_allocator_exit(void)
{
    int i = 0;

    for(i=0; i<WMA_STATIC_ALLOCATION_NUM; i++)
    {
        if(NULL != g_pMemGfpKernel[i])
        {
            kfree(g_pMemGfpKernel[i]);
            printk("wlanMemoryAllocator: Memory Freed. g_pMemGfpKernel[%d] = %p\n", i, g_pMemGfpKernel[i]);
        }
    }

    g_allocateCounter = 0;

    return;
}

//-------------------------------------------------------------
// Fake Allocation
// Return the pointers to allocated memory.
// g_allocateCounter will be incremented.
//-------------------------------------------------------------
void * static_malloc_kernel(size_t size)
{
    void * pRet = NULL;

    if(size > WMA_STATIC_ALLOCATION_SIZE)
    {
        printk("wlanMemoryAllocator: [ERROR] Param Size is over WMA_STATIC_ALLOCATION_SIZE.\n");
        return pRet;
    }

    if(g_allocateCounter < WMA_STATIC_ALLOCATION_NUM)
    {
        pRet = g_pMemGfpKernel[g_allocateCounter];
        memset(g_pMemGfpKernel[g_allocateCounter], 0, WMA_STATIC_ALLOCATION_SIZE);
        printk("wlanMemoryAllocator: Provide Allocated Memory. g_pMemGfpKernel[%d] = %p\n", g_allocateCounter, g_pMemGfpKernel[g_allocateCounter]);

        g_allocateCounter++;
    }
    else
    {
        printk("wlanMemoryAllocator: [ERROR] No More Allocated Memory.\n");
    }

    return pRet;
}
EXPORT_SYMBOL(static_malloc_kernel);

//-------------------------------------------------------------
// Fake Free
// This function decrements g_allocateCounter.
//-------------------------------------------------------------
void static_free_kernel(void * p)
{
    if(NULL == p)
    {
        printk("wlanMemoryAllocator: [ERROR] Param Pointer is NULL.\n");
        return;
    }

    // in this function, just decreament counter.
    if(0 < g_allocateCounter && g_allocateCounter <= WMA_STATIC_ALLOCATION_NUM)
    {
        g_allocateCounter --;
        printk("wlanMemoryAllocator: Allocated Memory Returned. g_allocateCounter = %d \n", g_allocateCounter);
    }
    else
    {
        printk("wlanMemoryAllocator: No Allocated Memory for Freeing. g_allocateCounter = %d \n", g_allocateCounter);
    }

    return;
}
EXPORT_SYMBOL(static_free_kernel);

module_init(wlan_allocator_init);
module_exit(wlan_allocator_exit);
